import datetime
import dateutil.parser as parser
from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.builder.builder import Builder
from usdm4.api.alias_code import AliasCode
from usdm4.api.code import Code


class Encoder:
    MODULE = "usdm4.encoder.encoder.Encoder"

    ZERO_DURATION = "PT0M"

    PHASE_MAP = [
        (
            ["0", "PRE-CLINICAL", "PRE CLINICAL"],
            {
                "code": "C54721",
                "m11_decode": "Early Phase 1",
                "cdisc_decode": "Phase 0 Trial",
            },
        ),
        (
            ["1", "I"],
            {
                "code": "C15600",
                "m11_decode": "Phase 1",
                "cdisc_decode": "Phase I Trial",
            },
        ),
        (
            ["1-2"],
            {
                "code": "C15693",
                "m11_decode": "Phase 1/Phase 2",
                "cdisc_decode": "Phase I/II Trial",
            },
        ),
        (
            ["1/2"],
            {
                "code": "C15693",
                "m11_decode": "Phase 1/Phase 2",
                "cdisc_decode": "Phase I/II Trial",
            },
        ),
        (
            ["1/2/3"],
            {
                "code": "C198366",
                "m11_decode": "Phase 1/Phase 2/Phase 3",
                "cdisc_decode": "Phase I/II/III Trial",
            },
        ),
        (
            ["1/3"],
            {
                "code": "C198367",
                "m11_decode": "Phase 1/Phase 3",
                "cdisc_decode": "Phase I/III Trial",
            },
        ),
        (
            ["1A", "IA"],
            {"code": "C199990", "m11_decode": None, "cdisc_decode": "Phase Ia Trial"},
        ),
        (
            ["1B", "IB"],
            {"code": "C199989", "m11_decode": None, "cdisc_decode": "Phase Ib Trial"},
        ),
        (
            ["2", "II"],
            {
                "code": "C15601",
                "m11_decode": "Phase 2",
                "cdisc_decode": "Phase II Trial",
            },
        ),
        (
            ["2-3", "II-III"],
            {
                "code": "C15694",
                "m11_decode": "Phase 2/Phase 3",
                "cdisc_decode": "Phase II/III Trial",
            },
        ),
        (
            ["2A", "IIA"],
            {"code": "C49686", "m11_decode": None, "cdisc_decode": "Phase IIa Trial"},
        ),
        (
            ["2B", "IIB"],
            {"code": "C49688", "m11_decode": None, "cdisc_decode": "Phase IIb Trial"},
        ),
        (
            ["3", "III"],
            {
                "code": "C15602",
                "m11_decode": "Phase 3",
                "cdisc_decode": "Phase III Trial",
            },
        ),
        (
            ["3A", "IIIA"],
            {"code": "C49687", "m11_decode": None, "cdisc_decode": "Phase IIIa Trial"},
        ),
        (
            ["3B", "IIIB"],
            {"code": "C49689", "m11_decode": None, "cdisc_decode": "Phase IIIb Trial"},
        ),
        (
            ["4", "IV"],
            {
                "code": "C15603",
                "m11_decode": "Phase 4",
                "cdisc_decode": "Phase IV Trial",
            },
        ),
        (
            ["5", "V"],
            {"code": "C47865", "m11_decode": None, "cdisc_decode": "Phase V Trial"},
        ),
        (
            ["2/3/4", "V"],
            {
                "code": "C217024",
                "m11_decode": "Phase 2/Phase 3/Phase 4",
                "cdisc_decode": None,
            },
        ),
    ]
    STATUS_MAP = [
        (["APPROVED"], {"code": "C25425", "decode": "Approved"}),
        (["DRAFT", "DFT"], {"code": "C85255", "decode": "Draft"}),
        (["FINAL"], {"code": "C25508", "decode": "Final"}),
        (["OBSOLETE"], {"code": "C63553", "decode": "Obsolete"}),
        (
            ["PENDING", "PRENDING REVIEW"],
            {"code": "C188862", "decode": "Pending Review"},
        ),
    ]
    REASON_MAP = [
        {"code": "C207612", "decode": "Regulatory Agency Request To Amend"},
        {"code": "C207608", "decode": "New Regulatory Guidance"},
        {"code": "C207605", "decode": "IRB/IEC Feedback"},
        {"code": "C207609", "decode": "New Safety Information Available"},
        {"code": "C207606", "decode": "Manufacturing Change"},
        {"code": "C207602", "decode": "IMP Addition"},
        {"code": "C207601", "decode": "Change In Strategy"},
        {"code": "C207600", "decode": "Change In Standard Of Care"},
        {
            "code": "C207607",
            "decode": "New Data Available (Other Than Safety Data)",
        },
        {"code": "C207604", "decode": "Investigator/Site Feedback"},
        {"code": "C207611", "decode": "Recruitment Difficulty"},
        {
            "code": "C207603",
            "decode": "Inconsistency And/Or Error In The Protocol",
        },
        {"code": "C207610", "decode": "Protocol Design Error"},
        {"code": "C17649", "decode": "Other"},
        {"code": "C48660", "decode": "Not Applicable"},
    ]

    BOOLEAN_MAP = {
        "true": True,
        "false": False,
        "1": True,
        "0": False,
        "yes": True,
        "no": False,
        "y": True,
        "n": False,
    }

    SCOPE_MAP = {
        "COUNTRY": {"code": "C25464", "decode": "Country"},
        "GLOBAL": {"code": "C68846", "decode": "Global"},
        "REGION": {"code": "C41129", "decode": "Region"},
    }

    def __init__(self, builder: Builder, errors: Errors):
        self._builder: Builder = builder
        self._errors: Errors = errors

    def phase(self, text: str) -> AliasCode:
        phase = text
        for word in ["PHASE", "TRIAL"]:
            phase = phase.upper().replace(word, "").strip() if phase else ""
        for tuple in self.PHASE_MAP:
            if phase in tuple[0]:
                entry = tuple[1]
                code = entry["code"]
                m11_decode = entry["m11_decode"]
                decode = m11_decode if m11_decode else entry["cdisc_decode"]
                phase_code = self._builder.cdisc_code(code, decode)
                self._errors.info(
                    f"Trial phase '{phase}' decoded as '{code}', '{decode}', {'using M11 decode' if m11_decode else 'using CDISC decode'}",
                    location=KlassMethodLocation(self.MODULE, "phase"),
                )
                if not m11_decode:
                    self._errors.warning(
                        f"Could not find M11 decode for phase '{phase}'"
                    )
                return self._builder.alias_code(phase_code)
        cdisc_phase_code = self._builder.cdisc_code(
            "C48660",
            "[Trial Phase] Not Applicable",
        )
        self._errors.warning(
            f"Trial phase '{phase}' not decoded",
            location=KlassMethodLocation(self.MODULE, "phase"),
        )
        return self._builder.alias_code(cdisc_phase_code)

    def document_status(self, text: str) -> Code:
        status = text.upper().strip() if text else ""
        for tuple in self.STATUS_MAP:
            if status in tuple[0]:
                entry = tuple[1]
                cdisc_code = self._builder.cdisc_code(
                    entry["code"],
                    entry["decode"],
                )
                self._errors.info(
                    f"Document status '{status}' decoded as '{entry['code']}', '{entry['decode']}'",
                    location=KlassMethodLocation(self.MODULE, "document_status"),
                )
                return cdisc_code
        cdisc_code = self._builder.cdisc_code("C85255", "Draft")
        self._errors.warning(
            f"Document status '{status}' not decoded",
            location=KlassMethodLocation(self.MODULE, "document_status"),
        )
        return cdisc_code

    def amendment_reason(self, reason_str: str):
        if reason_str:
            parts = reason_str.split(":")
            if len(parts) >= 2:
                reason_text = parts[1].strip()
                for reason in self.REASON_MAP:
                    if reason_text.upper() == reason["decode"].upper():
                        self._errors.info(
                            f"Amendment reason '{reason_text}' decoded as '{reason['code']}', '{reason['decode']}'"
                        )
                        code = self._builder.cdisc_code(
                            reason["code"], reason["decode"]
                        )
                        return {"code": code, "other_reason": ""}
            self._errors.warning(
                f"Unable to decode amendment reason '{reason_str}'",
                location=KlassMethodLocation(self.MODULE, "amendment_reason"),
            )
            code = self._builder.cdisc_code("C17649", "Other")
            return {"code": code, "other_reason": parts[-1].strip()}
        self._errors.warning(
            "Amendment reason not decoded, missing text",
            location=KlassMethodLocation(self.MODULE, "amendment_reason"),
        )
        code = self._builder.cdisc_code("C17649", "Other")
        return {"code": code, "other_reason": "No reason text found"}

    def geographic_scope(self, type: str) -> Code:
        if type.upper() in self.SCOPE_MAP:
            scope = self.SCOPE_MAP[type.upper()]
            return self._builder.cdisc_code(scope["code"], scope["decode"])
        self._errors.warning(
            f"Geographic scope not set for '{type}', setting global scope",
            location=KlassMethodLocation(self.MODULE, "geographic_scope"),
        )
        scope = self.SCOPE_MAP["GLOBAL"]
        return self._builder.cdisc_code(scope["code"], scope["decode"])

    def to_date(self, text: str) -> datetime.datetime | None:
        try:
            if text:
                input_text = text.strip()
                if input_text:
                    return parser.parse(input_text)
            return None
        except Exception as e:
            self._errors.exception(
                f"Failed to decode date text '{text}'",
                e,
                location=KlassMethodLocation(self.MODULE, "to_date"),
            )
            return None

    def iso8601_duration(self, value: int, unit: str) -> str:
        try:
            unit_text: str = unit.strip()
            if unit_text.upper() in ["Y", "YRS", "YR", "YEARS", "YEAR"]:
                return f"P{value}Y"
            if unit_text.upper() in ["MTHS", "MTH", "MONTHS", "MONTH"]:
                return f"P{value}M"
            if unit_text.upper() in ["W", "WKS", "WK", "WEEKS", "WEEK"]:
                return f"P{value}W"
            if unit_text.upper() in ["D", "DYS", "DY", "DAYS", "DAY"]:
                return f"P{value}D"
            if unit_text.upper() in ["H", "HRS", "HR", "HOURS", "HOUR"]:
                return f"PT{value}H"
            if unit_text.upper() in ["M", "MINS", "MIN", "MINUTES", "MINUTE"]:
                return f"PT{value}M"
            if unit_text.upper() in ["S", "SECS", "SEC", "SECONDS", "SECOND"]:
                return f"PT{value}S"
            self._errors.warning(
                f"Failed to encode ISO8601 duration of '{value}, {unit}'"
            )
            return self.ZERO_DURATION
        except Exception as e:
            self._errors.exception(
                f"Failed to encode ISO8601 duration of '{value}, {unit}'",
                e,
                location=KlassMethodLocation(self.MODULE, "iso8601_duration"),
            )
            return self.ZERO_DURATION

    def to_boolean(self, text: str) -> bool:
        return False if text is None else self.BOOLEAN_MAP.get(text.lower(), False)



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00066(RuleTemplate):
    """
    DDF00066: A scheduled decision instance is expected to refer to a default condition.

    Applies to: ScheduledDecisionInstance
    Attributes: defaultCondition
    """

    def __init__(self):
        super().__init__(
            "DDF00066",
            RuleTemplate.ERROR,
            "A scheduled decision instance is expected to refer to a default condition.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .alias_code import AliasCode


class GeographicScope(ApiBaseModelWithId):
    type: Code
    code: Union[AliasCode, None] = None
    instanceType: Literal["GeographicScope"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00060(RuleTemplate):
    """
    DDF00060: The value for each timing must be a non-negative duration specified in ISO 8601 format.

    Applies to: Timing
    Attributes: value
    """

    def __init__(self):
        super().__init__(
            "DDF00060",
            RuleTemplate.ERROR,
            "The value for each timing must be a non-negative duration specified in ISO 8601 format.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00110(RuleTemplate):
    """
    DDF00110: An eligibility criterion's category must be specified using the Category of Inclusion/Exclusion (C66797) SDTM codelist.

    Applies to: EligibilityCriterion
    Attributes: category
    """

    def __init__(self):
        super().__init__(
            "DDF00110",
            RuleTemplate.ERROR,
            "An eligibility criterion's category must be specified using the Category of Inclusion/Exclusion (C66797) SDTM codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0275(RuleTemplate):
    """
    CHK0275: If a biospecimen retention indicates that a type of biospecimen is retained, then there must be an indication of whether the type of biospecimen includes DNA.

    Applies to: BiospecimenRetention
    Attributes: includesDNA
    """

    def __init__(self):
        super().__init__(
            "CHK0275",
            RuleTemplate.ERROR,
            "If a biospecimen retention indicates that a type of biospecimen is retained, then there must be an indication of whether the type of biospecimen includes DNA.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00094(RuleTemplate):
    """
    DDF00094: Within a study version, if a date of a specific type exists with a global geographic scope then no other dates are expected with the same type.

    Applies to: StudyVersion
    Attributes: dateValues
    """

    def __init__(self):
        super().__init__(
            "DDF00094",
            RuleTemplate.ERROR,
            "Within a study version, if a date of a specific type exists with a global geographic scope then no other dates are expected with the same type.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import json
from typing import Union, List
from pydantic import BaseModel, Field
from .extension import ExtensionAttribute
from .serialize import serialize_as_json


class ApiBaseModel(BaseModel):
    def __init__(self, *args, **kwargs):
        kwargs["instanceType"] = self.__class__.__name__
        super().__init__(*args, **kwargs)

    def to_json(self):
        return json.dumps(self, default=serialize_as_json)


class ApiBaseModelWithIdOnly(ApiBaseModel):
    id: str = Field(min_length=1)

    def label_name(self) -> str:
        return ""


class ApiBaseModelWithId(ApiBaseModelWithIdOnly):
    extensionAttributes: List[ExtensionAttribute] = []

    def get_extension(self, url: str) -> ExtensionAttribute:
        return next(
            (x for x in self.extensionAttributes if x.url.upper() == url.upper()), None
        )


class ApiBaseModelWithIdAndDesc(ApiBaseModelWithId):
    description: Union[str, None] = None


class ApiBaseModelWithIdAndName(ApiBaseModelWithId):
    name: str = Field(min_length=1)

    def label_name(self) -> str:
        return self.name


class ApiBaseModelWithIdNameAndLabel(ApiBaseModelWithIdAndName):
    label: Union[str, None] = None

    def label_name(self) -> str:
        return self.label if self.label else self.name


class ApiBaseModelWithIdNameLabelAndDesc(ApiBaseModelWithIdNameAndLabel):
    description: Union[str, None] = None


class ApiBaseModelWithIdNameAndDesc(ApiBaseModelWithIdAndName):
    description: Union[str, None] = None



from typing import List, Literal, Union
from pydantic import Field
from .api_base_model import ApiBaseModel
from .study_definition_document import (
    StudyDefinitionDocument,
    StudyDefinitionDocumentVersion,
)
from .study_version import StudyVersion
from uuid import UUID


class Study(ApiBaseModel):
    id: Union[UUID, None] = None
    name: str = Field(min_length=1)
    description: Union[str, None] = None
    label: Union[str, None] = None
    versions: List[StudyVersion] = []
    documentedBy: List[StudyDefinitionDocument] = []
    instanceType: Literal["Study"]

    def document_by_template_name(self, name: str) -> StudyDefinitionDocument | None:
        return next(
            (x for x in self.documentedBy if x.templateName.upper() == name.upper()),
            None,
        )

    def document_templates(self) -> list[str]:
        return [x.templateName for x in self.documentedBy]

    def document_map(
        self,
    ) -> dict[str, dict[StudyDefinitionDocument, StudyDefinitionDocumentVersion]]:
        result = {}
        for doc in self.documentedBy:
            for version in doc.versions:
                result[version.id] = {"document": doc, "version": version}
        return result

    def first_version(self) -> StudyVersion | None:
        return self.versions[0] if len(self.versions) > 0 else None



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00092(RuleTemplate):
    """
    DDF00092: Date values associated to a study protocol document version must be unique regarding the combination of type and geographic scopes of the date.

    Applies to: StudyProtocolDocumentVersion
    Attributes: dateValues
    """

    def __init__(self):
        super().__init__(
            "DDF00092",
            RuleTemplate.ERROR,
            "Date values associated to a study protocol document version must be unique regarding the combination of type and geographic scopes of the date.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



class Library:
    def __init__(self, root_path: str):
        self.system = "ISO 639-1"
        self.version = "2007"

    def load(self) -> None:
        pass

    def code(self, decode: str) -> tuple[str | None, str | None]:
        return self._get_code(decode)

    def decode(self, code: str) -> tuple[str | None, str | None]:
        return self._get_decode(code)

    def code_or_decode(self, text: str) -> tuple[str | None, str | None]:
        code, name = self._get_decode(text)
        if not code:
            code, name = self._get_code(text)
        return code, name

    def _get_decode(self, code: str) -> tuple[str | None, str | None]:
        # Just support "en" for the moment
        return ("en", "English") if code == "en" else (None, None)

    def _get_code(self, decode: str) -> tuple[str | None, str | None]:
        # Just support "en" for the moment
        return ("en", "English") if decode.upper() == "ENGLISH" else (None, None)



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00133(RuleTemplate):
    """
    DDF00133: Within a study design, if a planned enrollment number is defined, it must be specified either in the study population or in all cohorts.

    Applies to: StudyDesignPopulation, StudyCohort
    Attributes: plannedEnrollmentNumber
    """

    def __init__(self):
        super().__init__(
            "DDF00133",
            RuleTemplate.ERROR,
            "Within a study design, if a planned enrollment number is defined, it must be specified either in the study population or in all cohorts.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import copy
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.builder.builder import Builder
from usdm4.api.address import Address
from usdm4.api.organization import Organization
from usdm4.api.identifier import StudyIdentifier
from usdm4.api.study_title import StudyTitle
from usdm4.api.study_role import StudyRole


from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation


class IdentificationAssembler(BaseAssembler):
    """
    Assembler for processing study identification information including titles, identifiers, and organizations.

    This assembler handles the creation of study titles, study identifiers, and associated organizations
    based on input data. It supports both standard predefined organizations (like CT.GOV, EMA, FDA)
    and custom non-standard organizations.

    Attributes:
        MODULE (str): Module path for error reporting
        TITLE_TYPES (list): Supported study title types
        TITLE_CODES (dict): Mapping of title types to CDISC codes
        ORG_CODES (dict): Mapping of organization types to CDISC codes
        STANDARD_ORGS (dict): Predefined standard organizations with their details
    """

    MODULE = "usdm4.assembler.identification_assembler.IdentificationAssembler"

    TITLE_TYPES = [
        "brief",
        "official",
        "public",
        "scientific",
        "acronym",
    ]

    ROLE_CODES = {
        "co-sponsor": {"code": "C215669", "decode": "Co-Sponsor"},
        "manufacturer": {"code": "C25392", "decode": "Manufacturer"},
        "investigator": {"code": "C25936", "decode": "Investigator"},
        "pharmacovigilance": {"code": "C215673", "decode": "Pharmacovigilance"},
        "project maanger": {"code": "C51851", "decode": "Project Manager"},
        "local sponsor": {"code": "C215670", "decode": "Local Sponsor"},
        "laboratory": {"code": "C37984", "decode": "Laboratory"},
        "study subject": {"code": "C41189", "decode": "Study Subject"},
        "medical expert": {"code": "C51876", "decode": "Medical Expert"},
        "statistician": {"code": "C51877", "decode": "Statistician"},
        "idmc": {"code": "C142578", "decode": "Independent Data Monitoring Committee"},
        "care provider": {"code": "C17445", "decode": "Care Provider"},
        "principal investigator": {
            "code": "C19924",
            "decode": "Principal investigator ",
        },
        "outcomes assessor": {"code": "C207599", "decode": "Outcomes Assessor      "},
        "dec": {"code": "C215671", "decode": "Dose Escalation Committee"},
        "clinical trial physician": {
            "code": "C215672",
            "decode": "Clinical Trial Physician",
        },
        "sponsor": {"code": "C70793", "decode": "Sponsor"},
        "adjudication Committee": {
            "code": "C78726",
            "decode": "Adjudication Committee",
        },
        "study site": {"code": "C80403", "decode": "Study Site"},
        "dsmb": {"code": "C142489", "decode": "Data Safety Monitoring Board"},
        "regulatory agency": {"code": "C188863", "decode": "Regulatory Agency"},
        "contract research": {"code": "C215662", "decode": "Contract Research"},
    }

    ROLE_ORGS = {
        "co_sponsor": {
            "type": "pharma",
            "role": "co-sponsor",
            "description": "The co-sponsor organization",
            "label": "",
            "identifier": "Not known",
            "identifierScheme": "Not known",
            "legalAddress": {},
        },
        "local_sponsor": {
            "type": "pharma",
            "role": "local sponsor",
            "description": "The local sponsor organization",
            "label": "",
            "identifier": "Not known",
            "identifierScheme": "Not known",
            "legalAddress": {},
        },
        "device_manufacturer": {
            "type": "medical_device",
            "role": "manufacturer",
            "description": "The medical device manufacturer",
            "label": "",
            "identifier": "Not known",
            "identifierScheme": "Not known",
            "legalAddress": {},
        },
    }

    TITLE_CODES = {
        "brief": {"code": "C207615", "decode": "Brief Study Title"},
        "official": {"code": "C207616", "decode": "Official Study Title"},
        "public": {"code": "C207617", "decode": "Public Study Title"},
        "scientific": {"code": "C207618", "decode": "Scientific Study Title"},
        "acronym": {"code": "C207646", "decode": "Study Acronym"},
    }

    ORG_CODES = {
        "registry": {"code": "C93453", "decode": "Clinical Study Registry"},
        "regulator": {"code": "C188863", "decode": "Regulatory Agency"},
        "healthcare": {"code": "C21541", "decode": "Healthcare Facility"},
        "pharma": {"code": "C54149", "decode": "Pharmaceutical Company"},
        "lab": {"code": "C37984", "decode": "Laboratory"},
        "cro": {"code": "C54148", "decode": "Contract Research Organization"},
        "gov": {"code": "C199144", "decode": "Government Institute"},
        "academic": {"code": "C18240", "decode": "Academic Institution"},
        "medical_device": {"code": "C215661", "decode": "Medical Device Company"},
    }
    STANDARD_ORGS = {
        "ct.gov": {
            "type": "registry",
            "role": None,
            "name": "CT.GOV",
            "label": "ClinicalTrials.gov",
            "description": "The US clinical trials registry",
            "identifier": "ClinicalTrials.gov",
            "identifierScheme": "National Institute of Health, US Government",
            "legalAddress": {
                "lines": ["National Library of Medicine", "8600 Rockville Pike"],
                "city": "Bethesda",
                "district": "",
                "state": "MD",
                "postalCode": "20894",
                "country": "USA",
            },
        },
        "ema": {
            "type": "regulator",
            "role": "regulatory agency",
            "name": "EMA",
            "label": "European Medicines Agency",
            "description": "The European medicines regulator",
            "identifier": "EMA",
            "identifierScheme": "European Union",
            "legalAddress": {
                "lines": ["Domenico Scarlattilaan 6"],
                "city": "Amsterdam",
                "district": "",
                "state": "",
                "postalCode": "1083 HS",
                "country": "NL",
            },
        },
        "fda": {
            "type": "regulator",
            "role": "regulatory agency",
            "name": "FDA",
            "label": "Food and Drug Administration",
            "identifier": "FDA",
            "description": "The US medicines regulator",
            "identifierScheme": "Health and Human Services, US Government",
            "legalAddress": {
                "lines": ["10903 New Hampshire Ave"],
                "city": "Silver Spring",
                "district": "",
                "state": "MD",
                "postalCode": "20903",
                "country": "USA",
            },
        },
    }

    def __init__(self, builder: Builder, errors: Errors):
        """
        Initialize the IdentificationAssembler.

        Args:
            builder (Builder): The builder instance for creating USDM objects
            errors (Errors): Error handling instance for logging issues
        """
        super().__init__(builder, errors)
        self.clear()

    def clear(self):
        self._titles = []
        self._organizations = []
        self._identifiers = []
        self._roles = []
        self._study_name = ""
        self._sponsor_signatory = None
        self._medical_expert = None
        self._compound_names = None
        self._compound_codes = None

    def execute(self, data: dict) -> None:
        """
        Process study identification data to create titles, identifiers, and organizations.

        This method processes the input data dictionary to create study titles, study identifiers,
        and associated organizations. It handles both standard predefined organizations and
        custom non-standard organizations.

        Args:
            data (dict): A complex dictionary containing study identification information with the following structure:
                {
                    "titles": {
                        "brief": str,           # Brief study title
                        "official": str,        # Official study title
                        "public": str,          # Public study title
                        "scientific": str,      # Scientific study title
                        "acronym": str,         # Study acronym
                    },
                    "identifiers": [            # List of study identifiers
                        {
                            "identifier": str,  # The actual identifier value
                            "scope": {          # Organization scope for the identifier
                                # Either use a standard predefined organization:
                                "standard": str,    # Key from STANDARD_ORGS (e.g., "ct.gov", "ema", "fda")

                                # OR define a custom non-standard organization:
                                "non_standard": {
                                    "type": str,            # Organization type (must match ORG_CODES keys)
                                    "role": str,            # Organization role (must match ROLE_CODES keys), can be None
                                    "name": str,            # Organization name
                                    "description": str,     # Organization description
                                    "label": str,           # Organization label/display name
                                    "identifier": str,      # Organization identifier
                                    "identifierScheme": str, # Scheme used for organization identifier
                                    "legalAddress": {       # Organization's legal address
                                        "lines": [str],     # Address lines (array of strings)
                                        "city": str,        # City name
                                        "district": str,    # District/region
                                        "state": str,       # State/province
                                        "postalCode": str,  # Postal/ZIP code
                                        "country": str      # ISO 3166 country code
                                    }
                                }
                            }
                        }
                    ],
                    "roles": {
                        "co_sponsor": {"name": [str], "legal_address": [address structure]},
                        "local_sponsor": {"name": [str], "legal_address": [address structure]},
                        "device_manufacturer": {"name": [str], "legal_address": [address structure]},
                    }
                }

        Note:
            - Title types must be one of: "brief", "official", "public", "scientific", "acronym"
            - Organization types for non_standard must match keys in ORG_CODES
            - Standard organization keys must match keys in STANDARD_ORGS
            - Country codes must be valid ISO 3166 codes
            - Each identifier must have either "standard" OR "non_standard" in scope, not both

        Raises:
            Various exceptions may be raised during object creation if data is invalid
        """
        # Make sure data ok.
        titles = data["titles"] if "titles" in data else {}
        identifiers = data["identifiers"] if "identifiers" in data else []
        roles = data["roles"] if "roles" in data else {}
        # Titles
        for type, text in titles.items():
            try:
                if type in self.TITLE_TYPES:
                    title = self._create_title(type, text)
                    if title:
                        self._titles.append(title)
                else:
                    self._errors.warning(
                        f"Title '{text}' of type '{type}' is not valid, ignored."
                    )
            except Exception as e:
                self._errors.exception(
                    f"Failed during creation of title '{text}' of type '{type}'",
                    e,
                    KlassMethodLocation(self.MODULE, "execute"),
                )

        # Identifiers
        id_details: dict
        for id_details in identifiers:
            try:
                scope = id_details["scope"]
                organization: dict = (
                    copy.deepcopy(self.STANDARD_ORGS[scope["standard"]])
                    if "standard" in scope
                    else scope["non_standard"]
                )

                # Address
                if organization["legalAddress"]:
                    organization["legalAddress"] = self._create_address(
                        organization["legalAddress"]
                    )

                # Identifier and scoping Organization
                org = self._create_organization(organization)
                if org:
                    identifier = self._create_identifier(id_details["identifier"], org)
                    if identifier:
                        self._organizations.append(org)
                        self._identifiers.append(identifier)
                    else:
                        self._errors.exception(
                            f"Failed to create identifier {id_details['identifier']} from organization '{organization}'",
                            KlassMethodLocation(self.MODULE, "execute"),
                        )
                else:
                    self._errors.exception(
                        f"Failed to create organization {organization}",
                        KlassMethodLocation(self.MODULE, "execute"),
                    )
            except Exception as e:
                self._errors.exception(
                    f"Failed during creation of identifier {id_details}",
                    e,
                    KlassMethodLocation(self.MODULE, "execute"),
                )
        for role, info in roles.items():
            try:
                if info is None:
                    continue
                organization = copy.deepcopy(self.ROLE_ORGS[role])
                # organization = self.ROLE_ORGS[role]
                organization["label"] = info["name"]
                organization["legalAddress"] = (
                    self._create_address(info["address"]) if "address" in info else None
                )
                org = self._create_organization(organization)
                if org:
                    self._organizations.append(org)
                else:
                    self._errors.exception(
                        f"Failed to create organization in {role}, {organization}",
                        KlassMethodLocation(self.MODULE, "execute"),
                    )
            except Exception as e:
                self._errors.exception(
                    f"Failed during creation of organization in {role}, {info}",
                    e,
                    KlassMethodLocation(self.MODULE, "execute"),
                )
        if "other" in data:
            self._sponsor_signatory = data["other"]["sponsor_signatory"]
            self._medical_expert = data["other"]["medical_expert"]
            self._compound_names = data["other"]["compound_names"]
            self._compound_codes = data["other"]["compound_codes"]

    @property
    def titles(self):
        return self._titles

    @property
    def organizations(self):
        return self._organizations

    @property
    def identifiers(self):
        return self._identifiers

    @property
    def roles(self):
        return self._roles

    @property
    def sponsor_signatory(self) -> str:
        return self._sponsor_signatory

    @property
    def medical_expert(self) -> str:
        return self._medical_expert

    @property
    def compound_names(self) -> str:
        return self._compound_names

    @property
    def compound_codes(self) -> str:
        return self._compound_codes

    def _create_address(self, address: dict) -> Address | None:
        try:
            self._errors.debug(
                f"Creating address, source data: {address}",
                KlassMethodLocation(self.MODULE, "_create_address"),
            )
            if ("country" in address) and address["country"]:
                address["country"] = self._builder.iso3166_code_or_decode(
                    address["country"]
                )
            else:
                address["country"] = None
            addr: Address = self._builder.create(Address, address)
            addr.set_text()
            self._errors.info(
                f"Address set to {addr.text}",
                KlassMethodLocation(self.MODULE, "_create_address"),
            )
            return addr
        except Exception as e:
            self._errors.exception(
                "Failed to create address object",
                e,
                KlassMethodLocation(self.MODULE, "_create_address"),
            )
            return None

    def _create_organization(self, organization: dict) -> Organization | None:
        try:
            role = None
            if organization["role"]:
                role: StudyRole = self._create_role(organization["role"])
                organization.pop("role")
            org_type = organization["type"]
            organization["type"] = self._builder.cdisc_code(
                self.ORG_CODES[org_type]["code"], self.ORG_CODES[org_type]["decode"]
            )
            organization["name"] = (
                organization["name"]
                if "name" in organization
                else self._label_to_name(organization["label"])
            )
            org = self._builder.create(Organization, organization)
            if role:
                role.organizationIds = [org.id]
            return org
        except Exception as e:
            self._errors.exception(
                "Failed during creation of organization",
                e,
                KlassMethodLocation(self.MODULE, "_create_organization"),
            )
            return None

    def _create_identifier(
        self, identifier: str, org: Organization
    ) -> StudyIdentifier | None:
        try:
            identifier = self._builder.create(
                StudyIdentifier, {"text": identifier, "scopeId": org.id}
            )
            return identifier
        except Exception as e:
            self._errors.exception(
                f"Failed during creation of identifier '{identifier}'",
                e,
                KlassMethodLocation(self.MODULE, "_create_identifier"),
            )
            return None

    def _create_title(self, type: str, text: str) -> StudyTitle | None:
        try:
            title_type = self._builder.cdisc_code(
                self.TITLE_CODES[type]["code"], self.TITLE_CODES[type]["decode"]
            )
            return self._builder.create(StudyTitle, {"text": text, "type": title_type})
        except Exception as e:
            self._errors.exception(
                f"Failed during creation of title '{text}'of type '{type}'",
                e,
                KlassMethodLocation(self.MODULE, "_create_title"),
            )
            return None

    def _create_role(self, type: str) -> StudyRole | None:
        try:
            role_type = self._builder.cdisc_code(
                self.ROLE_CODES[type]["code"], self.ROLE_CODES[type]["decode"]
            )
            index = len(self._roles)
            study_role: StudyRole = self._builder.create(
                StudyRole, {"name": f"ROLE_{index + 1}", "code": role_type}
            )
            if study_role:
                self._roles.append(study_role)
            return study_role
        except Exception as e:
            self._errors.exception(
                f"Failed during creation of role of type '{type}'",
                e,
                KlassMethodLocation(self.MODULE, "_create_role"),
            )
            return None
